(function (global) {
    "use strict";

    global.FRCC = global.FRCC || {};
    global.FRCC.AmeriCorps = global.FRCC.AmeriCorps || {};
    global.FRCC.AmeriCorps.SessionManagement = (function () {
        const SessionKeys = {
            Timeout: "UserTimeout",
            Pathname: "UserLocationPathname",
            Message: "ResetSessionMessage",
            NotificationAction: "ExtendSession",
            LastNotificationSent: "LastNotificationSent",
        };

        const Constants = {
            DebounceDelay: 1000,
            NotificationDelay: 60000,
            StaggerDelay: 10000,
        };

        const Thresholds = {
            SessionWarning: null,
            SessionWarningCountdown: null,
            SessionTimeout: null,
        };

        const Timers = {
            SessionWarning: null,
            SessionWarningCountdown: null,
            SessionTimeout: null,
        };

        let SessionChannel = null;
        let SessionNotificationsAllowed = false;
        let UserId = null;

        /*
            TO-DOs:
                Update other web files to call FRCC.AmeriCorps.SessionManagement.resetSessionTimers(); where appropriate
                    Update asyncAppAjax and _safeAjax to call resetSessionTimers
        */

        const checkForSessionTimeout = function (userId) {
            if (userId) {
                return;
            }

            const userTimeout = JSON.parse(getSessionStorageValue(SessionKeys.Timeout));
            if (!userTimeout) {
                return;
            }

            setSessionStorageValue(SessionKeys.Timeout, false);

            const returnUrl = getSessionStorageValue(SessionKeys.Pathname);
            FRCC.AmeriCorps.Modals.SessionManagement.openTimeoutModal(returnUrl);
        };

        const initializeSessionTimers = async function (userId) {
            if (!userId) {
                return;
            }
            UserId = userId;

            const sessionWarningThreshold = await FRCC.Portal.Utilities.getEYGAConfigurationSetting("eyfrcc_portalsessioninactivitywarningthreshold");
            const sessionTimeoutThreshold = await FRCC.Portal.Utilities.getEYGAConfigurationSetting("eyfrcc_portalsessioninactivitytimeoutthreshold");
            if (!sessionWarningThreshold || !sessionTimeoutThreshold) {
                return;
            }

            // Wait for reponse on Notifications API being granted/denied
            await initializeNotifications();

            // Create BroadcastChannel specific to the environment in case users have different environments open in the same browser
            SessionChannel = new BroadcastChannel(`${window.location.host}-sessionChannel`);

            Thresholds.SessionWarning = sessionWarningThreshold * 60 * 1000;
            Thresholds.SessionTimeout = sessionTimeoutThreshold * 60 * 1000;

            // TO-DO: Right now warning is hardcoded to start 1 minute from timeout. Make configurable?
            Thresholds.SessionWarningCountdown = Thresholds.SessionTimeout - 1 * 60 * 1000;

            resetSessionTimeout();
            attachHandlers();
        };

        const initializeNotifications = async function () {
            if (!("Notification" in window) || Notification.permission === "denied") {
                return;
            }

            if (Notification.permission === "granted") {
                SessionNotificationsAllowed = true;
                return;
            }

            Notification.requestPermission().then((permission) => {
                if (permission !== "granted") {
                    return;
                }
                SessionNotificationsAllowed = true;
            });
        };

        // Reset all session timeouts by broadcasting message
        const resetSessionTimeout = async function () {
            resetSessionTimers();
            SessionChannel.postMessage(SessionKeys.Message);
        };

        // Reset local session timers
        const resetSessionTimers = async function () {
            clearSessionTimers();
            setSessionTimers();
        };

        const clearSessionTimers = function () {
            Object.keys(Timers).forEach((timerKey) => {
                clearTimeout(Timers[timerKey]);
                Timers[timerKey] = null;
            });
        };

        const setSessionTimers = function () {
            Timers.SessionWarning = setTimeout(handleSessionWarning, Thresholds.SessionWarning);
            Timers.SessionWarningCountdown = setTimeout(handleSessionWarningCountdown, Thresholds.SessionWarningCountdown);
            Timers.SessionTimeout = setTimeout(handleSessionTimeout, Thresholds.SessionTimeout);
        };

        const handleSessionWarning = function () {
            FRCC.AmeriCorps.Modals.SessionManagement.openWarningModal(Thresholds.SessionWarning, Thresholds.SessionTimeout);

            if (!SessionNotificationsAllowed) {
                return;
            }

            // Multiple webpages could arrive at this point at the exact same time - Use random delay before checking if notification has been sent
            const delay = Math.random() * Constants.StaggerDelay;
            console.log(`Random delay (ms): ${delay}`);
            setTimeout(() => {
                if (!shouldSendNotification()) {
                    return;
                }

                createSessionNotification();
            }, delay);
        };

        const shouldSendNotification = function () {
            const notificationTime = getLocalStorageValue(SessionKeys.LastNotificationSent);
            const currentTime = new Date().getTime();

            // Check current time against last notification time - Allow if difference is greater than configured delay
            if (!notificationTime || currentTime - notificationTime > Constants.NotificationDelay) {
                setLocalStorageValue(SessionKeys.LastNotificationSent, currentTime);
                return true;
            }

            return false;
        };

        const createSessionNotification = function () {
            const notification = new Notification("Extend Session?", {
                body: "Your session will expire soon due to inactivity. All unsaved changes will be lost.",
            });

            notification.onclick = () => {
                extendSession();
                FRCC.AmeriCorps.Modals.SessionManagement.closeWarningModal();
            };
        };

        const handleSessionWarningCountdown = function () {
            FRCC.AmeriCorps.Modals.SessionManagement.displayCountdown();
        };

        const handleSessionTimeout = function (redirectToOrigin = false) {
            setSessionStorageValue(SessionKeys.Timeout, true);
            setSessionStorageValue(SessionKeys.Pathname, `${window.location.pathname}${window.location.search}`);

            if (!redirectToOrigin) {
                logout(); // Method stored on EYGA Header Web Template
                return;
            }

            window.location = window.location.origin;
        };

        let lock = false;
        let debounceTimer;
        const extendSession = async function () {
            if (lock) {
                return;
            }

            clearTimeout(debounceTimer);

            // Use debouncing to ensure repeated calls are not made against the api
            debounceTimer = setTimeout(async () => {
                lock = true;

                try {
                    await FRCC.Portal.Utilities.asyncAppAjax(
                        {
                            type: "GET",
                            url: `/_api/contacts(${UserId})?$select=contactid`,
                            contentType: "application/json",
                        },
                        true
                    );

                    resetSessionTimeout();
                } catch (error) {
                    console.error(error);

                    // User should always have read permission for their own contact
                    // If it is throwing an error, their session has timed out
                    if (error?.responseJSON?.error?.innererror?.type === "EntityPermissionReadIsMissing") {
                        handleSessionTimeout(true);
                    }
                } finally {
                    lock = false;
                }
            }, Constants.DebounceDelay);
        };

        const attachHandlers = function () {
            $(document).on("click", extendSession);
            $(document).on("keydown", extendSession);

            // Add event listener for BroadcastChannel - Reset local session timers when message is received
            SessionChannel.addEventListener("message", (e) => {
                if (e.data !== SessionKeys.Message) {
                    return;
                }

                resetSessionTimers();
                FRCC.AmeriCorps.Modals.SessionManagement.closeWarningModal();
            });
        };

        // Session Storage - will only be accessible while and by the window that created it is open (changes to data only affect current tab)
        const getSessionStorageValue = function (sessionKey) {
            return window.sessionStorage.getItem(sessionKey);
        };

        const setSessionStorageValue = function (sessionKey, value) {
            return window.sessionStorage.setItem(sessionKey, value);
        };

        // Local Storage - lasts until you delete it or the user deletes it (same data can be accessed by multiple tabs)
        const getLocalStorageValue = function (sessionKey) {
            return window.localStorage.getItem(sessionKey);
        };

        const setLocalStorageValue = function (sessionKey, value) {
            return window.localStorage.setItem(sessionKey, value);
        };

        return {
            checkForSessionTimeout: checkForSessionTimeout,
            initializeSessionTimers: initializeSessionTimers,
            extendSession: extendSession,
            getSessionStorageValue:getSessionStorageValue,
            setSessionStorageValue:setSessionStorageValue,
            getLocalStorageValue:getLocalStorageValue,
            setLocalStorageValue:setLocalStorageValue,
        };
    })();
})(this);
